import random

class TELSorting():
    '''
    WIP! Coming soon
    '''
    def __init__(self) -> None:
        pass

    def bubble_sort(self, arr: list[int | float]) -> list[int | float]:
        '''
        WIP! Coming soon
        '''
        n = len(arr)
        for i in range(n):
            for j in range(0, n-i-1):
                if arr[j] > arr[j+1]:
                    arr[j], arr[j+1] = arr[j+1], arr[j]
        return arr
    
    def selection_sort(self, arr: list[int | float]) -> list[int | float]:
        '''
        WIP! Coming soon
        '''
        n = len(arr)
        for i in range(n):
            min_index = i
            for j in range(i+1, n):
                if arr[j] < arr[min_index]:
                    min_index = j
            arr[i], arr[min_index] = arr[min_index], arr[i]
        return arr
    
    def insertion_sort(self, arr: list[int | float]) -> list[int | float]:
        '''
        WIP! Coming soon
        '''
        for i in range(1, len(arr)):
            key = arr[i]
            j = i - 1
            while j >= 0 and key < arr[j]:
                arr[j + 1] = arr[j]
                j -= 1
            arr[j + 1] = key
        return arr
    
    def quick_sort(self, arr: list[int | float]) -> list[int | float]:
        '''
        WIP! Coming soon
        '''
        if len(arr) <= 1:
            return arr
        pivot = arr[len(arr) // 2]
        left = [x for x in arr if x < pivot]
        middle = [x for x in arr if x == pivot]
        right = [x for x in arr if x > pivot]
        return self.quick_sort(left) + middle + self.quick_sort(right)
    
    def merge_sort(self, arr: list[int | float]) -> list[int | float]:
        '''
        WIP! Coming soon
        '''
        if len(arr) > 1:
            mid = len(arr) // 2
            left_half = arr[:mid]
            right_half = arr[mid:]

            self.merge_sort(left_half)
            self.merge_sort(right_half)

            i = j = k = 0
            while i < len(left_half) and j < len(right_half):
                if left_half[i] < right_half[j]:
                    arr[k] = left_half[i]
                    i += 1
                else:
                    arr[k] = right_half[j]
                    j += 1
                k += 1

            while i < len(left_half):
                arr[k] = left_half[i]
                i += 1
                k += 1

            while j < len(right_half):
                arr[k] = right_half[j]
                j += 1
                k += 1
            return arr
    
    def bogo_sort(self, arr: list[int | float]) -> list[int | float]:
        while not arr == sorted(arr):
            random.shuffle(arr)
        return arr

class TELSearching():
    '''
    WIP! Coming soon
    '''
    def __init__(self) -> None:
        pass

    def linear_search(self, arr: list[int | float], target: int | float):
        '''
        WIP! Coming soon
        '''
        for i in range(len(arr)):
            if arr[i] == target:
                return i
        return -1

    def binary_search(self, arr: list[int | float], target: int | float):
        '''
        WIP! Coming soon
        '''
        low, high = 0, len(arr) - 1
        while low <= high:
            mid = (low + high) // 2
            if arr[mid] == target:
                return mid
            elif arr[mid] < target:
                low = mid + 1
            else:
                high = mid - 1
        return -1
    
    def search(self, arr: list, target, comparison_func=None):
        '''
        WIP! Coming soon
        '''
        if comparison_func is None:
            comparison_func = lambda x, y: x == y

        for index, item in enumerate(arr):
            if comparison_func(item, target):
                return index
        return -1



from cryptography.hazmat.backends import default_backend
from cryptography.fernet import Cipher, algorithms, modes
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from cryptography.exceptions import InvalidTag

import os
import base64
import uuid
from typing import Tuple



class TELSymmetric():
    '''
    This is the main class for the symmetric encryption and deryption

    Arguments:
     - Iterations refer to the process of repeatedly applying a cryptographic algorithm a certain number of times to enhance security and complexity (Minimum: 100000)
     - Salt length determines the size of the random data added to plaintext before hashing or encrypting in password storage, enhancing security by preventing rainbow table attacks
     - The key length in encryption determines the size of the cryptographic key, directly influencing the security and strength of the encryption algorithm
    '''
    def __init__(self, iterations: int = 100000, salt_length: int = 32, key_length: int = 32) -> None:

        # Set iterations to minimum of 100000
        if iterations < 100000:
            self.iterations = 100000
        else:
            self.iterations = iterations
        
        if salt_length < 16:
            self.salt_length = 16
        else:
            self.salt_length = salt_length

        if key_length < 32:
            self.key_length = 32
        else:
            self.key_length = key_length

    def encrypt(self, plaintext: str, password: str) -> str:
        '''
        Encrypt the given text using a password to generate a key

        Arguments:
         - Plaintext is the text that is being encrypted
         - Password is the text that the encrpytion key is generated from

        Return:
         - Returns the encrypted text if no error occured

        Exeptions:
         - (ERROR) The standard error that is raised when something goes wrong
        '''

        try:
            plaintext = plaintext.encode()
            password = password.encode()

            salt = os.urandom(self.salt_length)

            # Generate a key using PBKDF2HMAC
            kdf = PBKDF2HMAC(
                algorithm=hashes.SHA384(),
                salt=salt,
                iterations=self.iterations,  # Use a larger iteration count for increased security
                length=self.key_length,
                backend=default_backend()
            )
            key = kdf.derive(password)

            # Generate a random nonce
            nonce = os.urandom(32) + uuid.uuid4().bytes + uuid.uuid4().bytes
            
            # Use AES-GCM for authenticated encryption
            cipher = Cipher(algorithms.AES(key), modes.GCM(nonce))
            encryptor = cipher.encryptor()
            ciphertext = encryptor.update(plaintext) + encryptor.finalize()

            # Serialize the nonce and tag with the ciphertext
            ciphertext_with_nonce_tag = encryptor.tag + ciphertext

            return base64.urlsafe_b64encode(salt + nonce + ciphertext_with_nonce_tag).decode()
        except Exception as msg:
            raise Exception(f"Error encrypting the text: {msg}")
        
    def decrypt(self, ciphertext: str, password: str) -> str:
        '''
        Encrypt the given text using a password to generate a key

        Arguments:
         - Ciphertext is the text that has been encrypted
         - Password is the text that the encrpytion key is generated from

        Return:
         - Returns the decrypted text if no error occured and the password is correct

        Exeptions:
         - (InvalidTag) This error is raised when the password is wrong or the ciphertext has been tampered with
         - (ERROR) The standard error that is raised when something goes wrong
        '''
        try:
            password = password.encode()

            # Decode the base64-encoded ciphertext
            ciphertext = base64.urlsafe_b64decode(ciphertext)

            # Extract the salt, nonce, and tag
            salt = ciphertext[:self.salt_length]
            nonce = ciphertext[self.salt_length:self.salt_length+64]
            tag = ciphertext[self.salt_length+64:self.salt_length+64+16]
            ciphertext = ciphertext[self.salt_length+64+16:]

            # Generate the same key using PBKDF2HMAC
            kdf = PBKDF2HMAC(
                algorithm=hashes.SHA384(),
                salt=salt,
                iterations=self.iterations,
                length=self.key_length,
                backend=default_backend()
            )
            key = kdf.derive(password)

            # Use AES-GCM for authenticated decryption
            cipher = Cipher(algorithms.AES(key), modes.GCM(nonce, tag))
            decryptor = cipher.decryptor()

            plaintext = decryptor.update(ciphertext) + decryptor.finalize()
            return plaintext.decode()
        except InvalidTag:
            raise InvalidTag("The password was wrong or the ciphertext has been tampered with")
        except Exception as msg:
            raise Exception(f"Error decrypting the text: {msg}")


















class TELAsymmetric():
    '''
    This is the main class for the asymmetric encryption and deryption pluss key management

    Arguments:
     - No arguments
    '''

    def __init__(self) -> None:
        pass

    def generate_key_pair(self) -> Tuple[rsa.RSAPrivateKey, rsa.RSAPublicKey]:
        '''
        Generates a key pair, one public key used for encrypting and a private key used for decryption

        Arguments:
         - No arguments

        Return:
         - Returns the key pair as Private key then Public key

        Exeptions:
         - (ERROR) The standard error that is raised when something goes wrong
        '''
        try:
            private_key = rsa.generate_private_key(
                public_exponent=65537,
                key_size=2048,
                backend=default_backend()
            )
            public_key = private_key.public_key()
            return (private_key, public_key)
        except Exception as msg:
            print(f"There was an error when generating the keys!\n{msg}")
            return "ERROR"

    def encrypt(self, plaintext: str, public_key: rsa.RSAPublicKey) -> str:
        '''
        Encrypt the given text using a given public key

        Arguments:
         - Plaintext is the text that is being encrypted
         - Public key used for encryption

        Return:
         - Returns the encrypted text if no error occured

        Exeptions:
         - (ERROR) The standard error that is raised when something goes wrong
        '''
        try:
            ciphertext = public_key.encrypt(
                plaintext.encode(),
                padding.OAEP(
                    mgf=padding.MGF1(algorithm=hashes.SHA512()),
                    algorithm=hashes.SHA512(),
                    label=None
                )
            )
            return ciphertext.hex()
        except Exception as msg:
            print(f"There was an error when encryption the text!\n{msg}")
            return "ERROR"

    def decrypt(self, ciphertext: str, private_key: rsa.RSAPrivateKey) -> str:
        '''
        Encrypt the given text using a password to generate a key

        Arguments:
         - Ciphertext is the text that has been encrypted
         - Private key used for the decryption of the ciphertext

        Return:
         - Returns the decrypted text if no error occured and the private key is correct

        Exeptions:
         - (ValueError) The given private key is wrong or the ciphertext has been tamperd with
         - (TypeError) The given private key is wrong or the ciphertext has been tamperd with
         - (ERROR) The standard error that is raised when something goes wrong
        '''
        try:
            plaintext = private_key.decrypt(
                bytes.fromhex(ciphertext),
                padding.OAEP(
                    mgf=padding.MGF1(algorithm=hashes.SHA512()),
                    algorithm=hashes.SHA512(),
                    label=None
                )
            )
            return plaintext.decode()
        except ValueError as msg:
            print(f"The ciphertext or key is wrong or not matching\n{msg}")
            return "ValueError"
        except TypeError as msg:
            print(f"The ciphertext or key is wrong or not matching\n{msg}")
            return "ValueError"
        except Exception as msg:
            print(f"There was an error when decrypting the text!\n{msg}")
            return "ERROR"
    
    def store_private_key(self, private_key: rsa.RSAPrivateKey, password: str, path: str, file_name: str = "private_key") -> None:
        '''
        Store the private key in a file encrypted and encoded with a password and Ascii85

        Arguments
         - Private key is the key that is stored
         - Password is the password used to encrypt and decrypt the private key
         - Path is the path to where the key are stored
         - File name is the name that the file with the key has

        Return:
         - No returns

        Exeptions:
         - (FileNotFoundError) The path or file name is wrong
         - (ERROR) The standard error that is raised when something goes wrong
        '''
        try:
            pem = private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.BestAvailableEncryption(password=password.encode())
            )
            b64_pem = base64.a85encode(pem, foldspaces=False, wrapcol=64, pad=True)
            with open(os.path.join(path, f'{file_name}.pem'), 'wb') as f:
                f.write(b64_pem)
        except FileNotFoundError as msg:
            raise FileNotFoundError(f"Error storing private key: {msg}")
        except Exception as msg:
            raise Exception(f"Error storing private key: {msg}")
    
    def store_public_key(self, public_key: rsa.RSAPrivateKey, path: str, file_name: str = "public_key") -> None:
        '''
        Store the public key in a file encoded in Ascii85

        Arguments:
         - Public key used for the encryption of the plaintext
         - Path is the path to where the key are stored
         - File name is the name that the file with the key has

        Return:
         - No returns

        Exeptions:
         - (FileNotFoundError) The path or file name is wrong
         - (ERROR) The standard error that is raised when something goes wrong
        '''
        try:
            pem = public_key.public_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PublicFormat.SubjectPublicKeyInfo
                )
            b64_pem = base64.a85encode(pem, foldspaces=False, wrapcol=64, pad=True)
            with open(os.path.join(path, f'{file_name}.pem'), 'wb') as f:
                f.write(b64_pem)
        except FileNotFoundError as msg:
            raise FileNotFoundError(f"Error storing public key: {msg}")
        except Exception as msg:
            raise Exception(f"Error storing public key: {msg}")
    
    def get_private_key(self, password: str, path: str, file_name: str = "private_key") -> rsa.RSAPrivateKey:
        '''
        Retrives a private key from a file and decrypts it using a password

        Arguments:
         - Password is the password used to encrypt and decrypt the private key
         - Path is the path to where the key are stored
         - File name is the name that the file with the key has

        Return:
         - No returns

        Exeptions:
         - (FileNotFoundError) The path or file name is wrong
         - (ValueError) The given password is incorrect
         - (ERROR) The standard error that is raised when something goes wrong
        '''
        try:
            with open(os.path.join(path, f'{file_name}.pem'), "rb") as f:
                key = base64.a85decode(f.read())
                private_key = serialization.load_pem_private_key(
                    key,
                    password=password.encode(),
                    backend=default_backend()
                )
            if private_key:
                return private_key
            else:
                raise Exception
        except FileNotFoundError as msg:
            raise FileNotFoundError(f"Error looking for private key: {msg}")
        except ValueError as msg:
            raise ValueError(f"The password is wrong: {msg}")
        except Exception as msg:
            raise Exception(f"Error getting private key: {msg}")
    
    def get_public_key(self, path: str, file_name: str = "public_key") -> rsa.RSAPublicKey:
        '''
        Retrives a private key from a file and decrypts it using a password

        Arguments:
         - Password is the password used to encrypt and decrypt the private key
         - Path is the path to where the key are stored
         - File name is the name that the file with the key has

        Return:
         - No returns

        Exeptions:
         - (FileNotFoundError) The path or file name is wrong
         - (ERROR) The standard error that is raised when something goes wrong
        '''
        try:
            with open(os.path.join(path, f'{file_name}.pem'), "rb") as f:
                key = base64.a85decode(f.read())
                public_key = serialization.load_pem_public_key(
                    key,
                    backend=default_backend()
                )
            if public_key:
                return public_key
            else:
                raise Exception
        except FileNotFoundError as msg:
            raise FileNotFoundError(f"Error looking for public key: {msg}")
        except Exception as msg:
            raise Exception(f"Error getting public key: {msg}")


import os
import shutil

class TELFileManager:
    def create_directory(self, dir: str) -> str:
        try:
            os.makedirs(dir, exist_ok=True)
            return dir
        except OSError as msg:
            raise OSError(f"Error creating directory: {msg}")
        except Exception as msg:
            raise Exception(f"Error deleting directory: {msg}")

    def delete_directory(self, dir: str) -> bool:
        try:
            if os.path.exists(dir) and os.path.isdir(dir):
                shutil.rmtree(dir)
                return True
            return False
        except Exception as msg:
            raise Exception(f"Error deleting directory: {msg}")

    def list_files(self, dir: str) -> list[str]:
        try:
            if os.path.exists(dir) and os.path.isdir(dir):
                files = [f for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))]
                return files
            return []
        except Exception as msg:
            raise Exception(f"Error listing files: {msg}")

    def copy_file(self, src_file: str, dest_dir: str) -> bool:
        dest_path = os.path.join(dest_dir, src_file)
        try:
            if os.path.exists(src_file) and os.path.isfile(src_file):
                shutil.copy(src_file, dest_path)
                return True
            return False
        except Exception as msg:
            raise Exception(f"Error copying file: {msg}")

    def move_file(self, src_file: str, dest_dir: str) -> bool:
        dest_path = os.path.join(dest_dir, src_file)
        try:
            if os.path.exists(src_file) and os.path.isfile(src_file):
                shutil.move(src_file, dest_path)
                return True
            return False
        except Exception as msg:
            raise Exception(f"Error moving file: {msg}")



import re

class TELRegex:
    '''
    WIP! Coming soon
    '''
    def __init__(self, default_pattern):
        self.default_pattern = default_pattern

        self.common_patterns = {
            "email": r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b',
            "url": r'https?://(?:www\.)?[A-Za-z0-9.-]+(?:\.[A-Za-z]{2,})+[\w/]*',
            "date": r'\d{4}-\d{2}-\d{2}',
            "phone": r'\b(?:\+\d{1,2}\s?)?(?:\(\d{3}\)|\d{3})[-.\s]?\d{3}[-.\s]?\d{4}\b',
            "ip": r'\b(?:\d{1,3}\.){3}\d{1,3}\b',
            "color": r'#[0-9A-Fa-f]{6}',
            "word": r'\b\w+\b',
            "text": r'[a-z]+',
            "number": r'\d+(?:\.\d+)?'
        }
    
    def pattern(self, name: str):
        '''
        WIP! Coming soon
        '''
        try:
            return self.common_patterns[name.lower()]
        except KeyError:
            raise KeyError(f'The name "{name}" is not a valid name. Please choose a name from this list: email, url, date, phone, ip, color, word, text, number')

    def match(self, text, pattern=None):
        '''
        WIP! Coming soon
        '''
        pattern_to_use = pattern if pattern is not None else self.default_pattern
        return re.match(pattern_to_use, text)

    def search(self, text, pattern=None):
        '''
        WIP! Coming soon
        '''
        pattern_to_use = pattern if pattern is not None else self.default_pattern
        return re.search(pattern_to_use, text)

    def findall(self, text, pattern=None):
        '''
        WIP! Coming soon
        '''
        pattern_to_use = pattern if pattern is not None else self.default_pattern
        return re.findall(pattern_to_use, text)

    def finditer(self, text, pattern=None):
        '''
        WIP! Coming soon
        '''
        pattern_to_use = pattern if pattern is not None else self.default_pattern
        return re.finditer(pattern_to_use, text)

    def replace(self, text, replacement, pattern=None):
        '''
        WIP! Coming soon
        '''
        pattern_to_use = pattern if pattern is not None else self.default_pattern
        return re.sub(pattern_to_use, replacement, text)

    @staticmethod
    def escape(text):
        return re.escape(text)

    @staticmethod
    def split(text, pattern):
        return re.split(pattern, text)



class TELMath():
    '''
    WIP! Coming soon
    '''
    def __init__(self) -> None:
        pass

    def add(self, x: int | float | str, y: int | float | str) -> float:
        '''
        WIP! Coming soon
        '''
        return float(x + y)

    def add_list(self, arr: list[int | float | str]) -> float:
        '''
        WIP! Coming soon
        '''
        result = 0.0
        clean_numbers = []
        for num in arr:
            if type(num) == str:
                clean_numbers.append(''.join(char for char in num if char.isdigit()))
            else:
                clean_numbers.append(num)
        if type(clean_numbers) == list:
            for num in clean_numbers:
                try:
                    result += float(num)
                except ValueError as msg:
                    print(f"{num} is a {type(num)} which is not able to be added!\n{msg}")
                except Exception as msg:
                    print(f"An error occured!\n{msg}")
            return float(result)
        else:
            return float(result)
    
    def subtract(self, x: int | float | str, y: int | float | str) -> float:
        '''
        WIP! Coming soon
        '''
        return float(x - y)

    def subtract_list(self, arr_x: list[int | float | str], arr_y: list[int | float | str]) -> float:
        '''
        WIP! Coming soon
        '''
        x = self.add_list(arr=arr_x)
        y = self.add_list(arr=arr_y)

        try:
            return float(x - y)
        except ValueError as msg:
            print(f"{x} or {y} is a {type(x)} or {type(y)} which is not able to be subtracted!\n{msg}")
        except Exception as msg:
            print(f"An error occured!\n{msg}")
    
    def average(self, arr: list[int | float | str]) -> float:
        '''
        WIP! Coming soon
        '''
        length = len(arr)
        _sum = self.add_list(arr=arr)

        try:
            return float(_sum / length)
        except ValueError as msg:
            print(f"{_sum} is not able to be averaged!\n{msg}")
        except Exception as msg:
            print(f"An error occured!\n{msg}")
    
    def median(self, arr: list[int | float | str]) -> float:
        '''
        WIP! Coming soon
        '''
        sorted_numbers = sorted(arr)
        length = len(sorted_numbers)
        
        if length % 2 == 1:
            median = sorted_numbers[length // 2]
        else:
            middle1 = sorted_numbers[length // 2 - 1]
            middle2 = sorted_numbers[length // 2]
            median = (middle1 + middle2) / 2
        
        return median


import base64

class TELEncoding():
    '''
    WIP! Coming soon
    '''
    def __init__(self) -> None:
        pass

    def base64_encode(self, plaintext: str, encoding: str = 'utf-8') -> str:
        '''
        WIP! Coming soon
        '''
        try:
            encoded_bytes = base64.b64encode(plaintext.encode(encoding=encoding))
            encoded_string = encoded_bytes.decode(encoding=encoding)
            return encoded_string
        except UnicodeEncodeError as msg:
            raise UnicodeEncodeError(f"There was an error when encoding the text: {msg}")
        except UnicodeDecodeError as msg:
            raise UnicodeDecodeError(f"There was an error when decoding the text: {msg}")
        except Exception as msg:
            raise Exception(f"There was an error when encoding the text: {msg}")
    
    def base64_decode(self, encoded_text: str, encoding: str = 'utf-8') -> str:
        '''
        WIP! Coming soon
        '''
        try:
            decoded_bytes = base64.b64decode(encoded_text.encode(encoding=encoding))
            decoded_string = decoded_bytes.decode(encoding=encoding)
            return decoded_string
        except UnicodeEncodeError as msg:
            raise UnicodeEncodeError(f"There was an error when encoding the text: {msg}")
        except UnicodeDecodeError as msg:
            raise UnicodeDecodeError(f"There was an error when decoding the text: {msg}")
        except Exception as msg:
            raise Exception(f"There was an error when encoding the text: {msg}")

    def hex_encode(self, plaintext: str, encoding: str = 'utf-8') -> str:
        '''
        WIP! Coming soon
        '''
        try:
            encoded_string = plaintext.encode(encoding=encoding).hex()
            return encoded_string
        except UnicodeEncodeError as msg:
            raise UnicodeEncodeError(f"There was an error when encoding the text: {msg}")
        except UnicodeDecodeError as msg:
            raise UnicodeDecodeError(f"There was an error when decoding the text: {msg}")
        except Exception as msg:
            raise Exception(f"There was an error when encoding the text: {msg}")
    
    def hex_decode(self, encoded_text: str, encoding: str = 'utf-8') -> str:
        '''
        WIP! Coming soon
        '''
        try:
            decoded_string = bytes.fromhex(encoded_text).decode(encoding=encoding)
            return decoded_string
        except UnicodeEncodeError as msg:
            raise UnicodeEncodeError(f"There was an error when encoding the text: {msg}")
        except UnicodeDecodeError as msg:
            raise UnicodeDecodeError(f"There was an error when decoding the text: {msg}")
        except Exception as msg:
            raise Exception(f"There was an error when encoding the text: {msg}")
        
    def binary_encode(self, plaintext: str, as_bytes: bool = False) -> str | bytes:
        '''
        WIP! Coming soon
        '''
        try:
            if as_bytes:
                encoded_bytes = bytes([ord(char) for char in plaintext])
            else:
                encoded_bytes = ' '.join(format(ord(char), '08b') for char in plaintext)
            return encoded_bytes
        except UnicodeEncodeError as msg:
            raise UnicodeEncodeError(f"There was an error when encoding the text: {msg}")
        except UnicodeDecodeError as msg:
            raise UnicodeDecodeError(f"There was an error when decoding the text: {msg}")
        except Exception as msg:
            raise Exception(f"There was an error when encoding the text: {msg}")
    
    def binary_decode(self, encoded_text: str | bytes) -> str:
        '''
        WIP! Coming soon
        '''
        try:
            if type(encoded_text) == str:
                binary_list = encoded_text.split()
                decoded_string = ''.join(chr(int(binary, 2)) for binary in binary_list)
            elif type(encoded_text) == bytes:
                decoded_string = ''.join(chr(byte) for byte in encoded_text)
            return decoded_string
        except UnicodeEncodeError as msg:
            raise UnicodeEncodeError(f"There was an error when encoding the text: {msg}")
        except UnicodeDecodeError as msg:
            raise UnicodeDecodeError(f"There was an error when decoding the text: {msg}")
        except Exception as msg:
            raise Exception(f"There was an error when encoding the text: {msg}")
